@@ -17,7 +17,7 @@
 // Copyright (C) 2007 Iñigo Martínez <inigomartinez@gmail.com>
 // Copyright (C) 2008 Brad Hards <bradh@kde.org>
 // Copyright (C) 2008, 2010 Carlos Garcia Campos <carlosgc@gnome.org>
-// Copyright (C) 2009-2013, 2017, 2018, 2021, 2024 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2009-2013, 2017, 2018, 2021, 2024, 2025 Albert Astals Cid <aacid@kde.org>
 // Copyright (C) 2009, 2010, 2012, 2013 Thomas Freitag <Thomas.Freitag@alfa.de>
 // Copyright (C) 2010 David Benjamin <davidben@mit.edu>
 // Copyright (C) 2010 Christian Feuersänger <cfeuersaenger@googlemail.com>
@@ -41,6 +41,7 @@
 #include "Object.h"
 #include "PopplerCache.h"
 
+#include <stack>
 #include <vector>
 
 class GooString;
@@ -169,7 +170,13 @@ public:
     XRef *getXRef() { return xref; }
 
     // Interpret a stream or array of streams.
-    void display(Object *obj, bool topLevel = true);
+    enum class DisplayType
+    {
+        TopLevel,
+        Type3Font,
+        Form
+    };
+    void display(Object *obj, DisplayType displayType = DisplayType::TopLevel);
 
     // Display an annotation, given its appearance (a Form XObject),
     // border style, and bounding box (in default user space).
@@ -192,8 +199,8 @@ public:
 
     bool checkTransparencyGroup(Dict *resDict);
 
-    void drawForm(Object *str, Dict *resDict, const double *matrix, const double *bbox, bool transpGroup = false, bool softMask = false, GfxColorSpace *blendingColorSpace = nullptr, bool isolated = false, bool knockout = false,
-                  bool alpha = false, Function *transferFunc = nullptr, GfxColor *backdropColor = nullptr);
+    void drawForm(Object *str, Dict *resDict, const std::array<double, 6> &matrix, const std::array<double, 4> &bbox, bool transpGroup = false, bool softMask = false, GfxColorSpace *blendingColorSpace = nullptr, bool isolated = false,
+                  bool knockout = false, bool alpha = false, Function *transferFunc = nullptr, GfxColor *backdropColor = nullptr);
 
     void pushResources(Dict *resDict);
     void popResources();
@@ -210,6 +217,9 @@ private:
     GfxResources *res; // resource stack
     int updateLevel;
 
+    std::stack<DisplayType> displayTypes;
+    std::stack<bool> type3FontIsD1;
+
     GfxState *state; // current graphics state
     int stackHeight; // the height of the current graphics stack
     std::vector<int> stateGuards; // a stack of state limits; to guard against unmatched pops
@@ -235,7 +245,7 @@ private:
 
     static const Operator opTab[]; // table of operators
 
-    void go(bool topLevel);
+    void go(DisplayType displayType);
     void execOp(Object *cmd, Object args[], int numArgs);
     const Operator *findOp(const char *name);
     bool checkArg(Object *arg, TchkType type);
